import logging
from collections.abc import Iterable

import orjson
from anystore.io import smart_open, smart_stream
from banal import is_listish
from nomenklatura.entity import CE, CompositeEntity
from nomenklatura.util import PathLike

from ftmq.query import Query
from ftmq.store import Store, get_store
from ftmq.types import CEGenerator, SDict
from ftmq.util import get_statements, make_dataset, make_proxy

log = logging.getLogger(__name__)

DEFAULT_MODE = "rb"


def smart_get_store(uri: PathLike, **kwargs) -> Store | None:
    try:
        return get_store(uri, **kwargs)
    except NotImplementedError:
        return


def smart_read_proxies(
    uri: PathLike | Iterable[PathLike],
    mode: str | None = DEFAULT_MODE,
    serialize: bool | None = True,
    query: Query | None = None,
    **store_kwargs,
) -> CEGenerator:
    if is_listish(uri):
        for u in uri:
            yield from smart_read_proxies(u, mode, serialize, query)
        return

    store = smart_get_store(uri, **store_kwargs)
    if store is not None:
        view = store.query()
        yield from view.entities(query)
        return

    lines = smart_stream(uri)
    lines = (orjson.loads(line) for line in lines)
    if serialize or query:
        q = query or Query()
        proxies = (make_proxy(line) for line in lines)
        yield from q.apply_iter(proxies)
    else:
        yield from lines


def smart_write_proxies(
    uri: PathLike,
    proxies: Iterable[CE | SDict],
    mode: str | None = "wb",
    serialize: bool | None = False,
    **store_kwargs,
) -> int:
    ix = 0
    if proxies is None:  # FIXME how could this happen
        return ix

    store = smart_get_store(uri, **store_kwargs)
    if store is not None:
        dataset = store_kwargs.get("dataset")
        if dataset is not None:
            proxies = apply_datasets(proxies, dataset, replace=True)
        with store.writer() as bulk:
            for proxy in proxies:
                ix += 1
                bulk.add_entity(proxy)
                if ix % 1_000 == 0:
                    log.info("Writing proxy %d ..." % ix)
        return ix

    with smart_open(uri, mode=mode) as fh:
        for proxy in proxies:
            ix += 1
            if serialize:
                proxy = proxy.to_dict()
            fh.write(orjson.dumps(proxy, option=orjson.OPT_APPEND_NEWLINE))
    return ix


def apply_datasets(
    proxies: Iterable[CE], *datasets: Iterable[str], replace: bool | None = False
) -> CEGenerator:
    for proxy in proxies:
        if datasets:
            if not replace:
                datasets = proxy.datasets | set(datasets)
            statements = get_statements(proxy, *datasets)
            dataset = make_dataset(list(datasets)[0])
        yield CompositeEntity.from_statements(dataset, statements)



import logging
from typing import Iterable

from nomenklatura import store as nk
from nomenklatura.resolver import Resolver

from ftmq.aggregations import AggregatorResult
from ftmq.model.coverage import Collector, DatasetStats
from ftmq.model.dataset import C, Dataset
from ftmq.query import Q
from ftmq.types import CE, CEGenerator
from ftmq.util import DefaultDataset, make_dataset

log = logging.getLogger(__name__)


class Store(nk.Store):
    def __init__(
        self,
        catalog: C | None = None,
        dataset: Dataset | str | None = None,
        resolver: Resolver | None = None,
        **kwargs,
    ) -> None:
        if dataset is not None:
            if isinstance(dataset, str):
                dataset = Dataset(name=dataset)
            dataset = make_dataset(dataset.name)
        elif catalog is not None:
            dataset = catalog.get_scope()
        else:
            dataset = DefaultDataset
        super().__init__(dataset=dataset, resolver=resolver or Resolver(), **kwargs)

    def get_catalog(self) -> C:
        # return implicit catalog computed from current datasets in store
        raise NotImplementedError

    def iterate(self) -> CEGenerator:
        catalog = self.get_catalog()
        view = self.view(catalog.get_scope())
        yield from view.entities()

    def resolve(self, dataset: str | Dataset | None = None) -> None:
        if not self.resolver.edges:
            return
        if dataset is not None:
            if isinstance(dataset, str):
                dataset = make_dataset(dataset)
            elif isinstance(dataset, Dataset):
                dataset = make_dataset(dataset.name)
            view = self.view(scope=dataset)
            entities = view.entities()
        else:
            entities = self.iterate()
        for ix, entity in enumerate(entities):
            if entity.id in self.resolver.nodes:
                self.update(self.resolver.get_canonical(entity.id))
            if ix and ix % 10_000 == 0:
                log.info("Resolving entity %d ..." % ix)


class View(nk.base.View):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache = {}

    def entities(self, query: Q | None = None) -> CEGenerator:
        view = self.store.view(self.scope)
        if query:
            yield from query.apply_iter(view.entities())
        else:
            yield from view.entities()

    def get_adjacents(self, proxies: Iterable[CE]) -> set[CE]:
        seen: set[CE] = set()
        for proxy in proxies:
            for _, adjacent in self.get_adjacent(proxy):
                if adjacent.id not in seen:
                    seen.add(adjacent)
        return seen

    def stats(self, query: Q | None = None) -> DatasetStats:
        key = f"stats-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        c = Collector()
        cov = c.collect_many(self.entities(query))
        self._cache[key] = cov
        return cov

    def aggregations(self, query: Q) -> AggregatorResult | None:
        if not query.aggregations:
            return
        key = f"agg-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        _ = [x for x in self.entities(query)]
        res = dict(query.aggregator.result)
        self._cache[key] = res
        return res



from collections import Counter
from typing import Any

from followthemoney import model

from ftmq.enums import Properties
from ftmq.model.mixins import BaseModel
from ftmq.types import CE, CEGenerator, DateLike, Frequencies
from ftmq.util import get_country_name


class Schema(BaseModel):
    name: str
    count: int
    label: str
    plural: str

    def __init__(self, **data):
        schema = model.get(data["name"])
        data["label"] = schema.label
        data["plural"] = schema.plural
        super().__init__(**data)


class Country(BaseModel):
    code: str
    count: int
    label: str | None = None

    def __init__(self, **data):
        data["label"] = get_country_name(data["code"])
        super().__init__(**data)


class Schemata(BaseModel):
    total: int = 0
    countries: list[Country] | None = []
    schemata: list[Schema] | None = []


class Coverage(BaseModel):
    start: DateLike | None = None
    end: DateLike | None = None
    frequency: Frequencies | None = "unknown"
    countries: list[str] | None = []
    schedule: str | None = None

    @property
    def years(self) -> tuple[int | None, int | None]:
        """
        Return min / max year extend
        """
        return (
            self.start.year if self.start else None,
            self.end.year if self.end else None,
        )


class DatasetStats(BaseModel):
    coverage: Coverage
    things: Schemata
    intervals: Schemata
    entity_count: int | None = 0


class Collector:
    def __init__(self):
        self.things = Counter()
        self.things_countries = Counter()
        self.intervals = Counter()
        self.intervals_countries = Counter()
        self.start = set()
        self.end = set()

    def collect(self, proxy: CE) -> None:
        if proxy.schema.is_a("Thing"):
            self.things[proxy.schema.name] += 1
            for country in proxy.countries:
                self.things_countries[country] += 1
        else:
            self.intervals[proxy.schema.name] += 1
            for country in proxy.countries:
                self.intervals_countries[country] += 1
        self.start.update(proxy.get(Properties.startDate, quiet=True))
        self.start.update(proxy.get(Properties.date, quiet=True))
        self.end.update(proxy.get(Properties.endDate, quiet=True))
        self.end.update(proxy.get(Properties.date, quiet=True))

    def export(self) -> DatasetStats:
        start = min(self.start) if self.start else None
        end = max(self.end) if self.end else None
        countries = set(self.things_countries.keys()) | set(
            self.intervals_countries.keys()
        )
        coverage = Coverage(start=start, end=end, countries=countries)
        things = Schemata(
            schemata=[Schema(name=k, count=v) for k, v in self.things.items()],
            countries=[
                Country(code=k, count=v) for k, v in self.things_countries.items()
            ],
            total=self.things.total(),
        )
        intervals = Schemata(
            schemata=[Schema(name=k, count=v) for k, v in self.intervals.items()],
            countries=[
                Country(code=k, count=v) for k, v in self.intervals_countries.items()
            ],
            total=self.intervals.total(),
        )
        return DatasetStats(
            coverage=coverage,
            things=things,
            intervals=intervals,
            entity_count=things.total + intervals.total,
        )

    def to_dict(self) -> dict[str, Any]:
        data = self.export()
        return data.model_dump()

    def apply(self, proxies: CEGenerator) -> CEGenerator:
        """
        Generate coverage from an input stream of proxies
        This returns a generator again, so actual collection of coverage stats
        will happen if the actual generator is executed
        """
        for proxy in proxies:
            self.collect(proxy)
            yield proxy

    def collect_many(self, proxies: CEGenerator) -> DatasetStats:
        for proxy in proxies:
            self.collect(proxy)
        return self.export()
